﻿Imports Microsoft.VisualBasic
Imports System
' demo of Fisher's Linear Discriminate Analysis (LDA)
' for binary classification
' note: a different LDA is Latent Dirichlet Allocation 

Namespace LinearDiscriminate
    Friend Class LinearDiscriminateProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(VBCrLf & "Begin linear discriminate analysis (LDA) demo" & VBCrLf)
            Console.WriteLine("Goal is to predict Politic (liberal = 0 or conservative = 1) ")
            Console.WriteLine("from person's Age and Income" & VBCrLf)

            Dim data(7)() As Double
            data(0) = New Double() {1.0, 4.0, 0}
            data(1) = New Double() {2.0, 2.0, 0}
            data(2) = New Double() {3.0, 3.0, 0}

            data(3) = New Double() {6.0, 6.0, 1}
            data(4) = New Double() {7.0, 7.0, 1}
            data(5) = New Double() {8.0, 9.0, 1}
            data(6) = New Double() {8.0, 7.0, 1}
            data(7) = New Double() {9.0, 8.0, 1}

            Console.WriteLine("Normalized and encoded training data:" & VBCrLf)
            Console.WriteLine(" Age   Income  Politic")
            Console.WriteLine("----------------------")
            ShowMatrix(data, 2, True)

            Console.WriteLine(VBCrLf & "Calculating the discriminate w")
            Dim w()() As Double = Discriminate(data, True) ' calls ScatterWithin which calls Scatter
            Console.WriteLine("Done calculating discriminate. Discriminate is: " & VBCrLf)
            ShowMatrix(w, 6, True)

            Dim x() As Double = {4.0, 5.0}
            Console.WriteLine("Predicting class for Age Income =")
            ShowVector(x, 1)
            Dim c As Integer = Prediction(data, x, w)
            Console.Write(VBCrLf & "Predicted class: " & c)
            If c = 0 Then
                Console.WriteLine(" (liberal)")
            Else
                Console.WriteLine(" (conservative)")
            End If

            ' 2. pooled covariance approach to compute Discriminate
            'Dim pooled()() As Double = pooled(data, False)
            'Dim pooledInv()() As Double = MatrixInverse(pooled)
            'Dim M0()() As Double = Mean(data, 0)
            'Dim M1()() As Double = Mean(data, 1)
            'Dim diff()() As Double = MatrixSubtract(M0, M1)
            'Dim prod()() As Double = MatrixProduct(pooledInv, diff)
            'Dim pooledW()() As Double = Unitize(prod)
            'ShowMatrix(pooledW, 6, True) ' exactly same as non-pooled scatter
            'Console.WriteLine(VBCrLf & "Discriminate calculated using pooled covariance:")
            'ShowMatrix(pooledW, 4, True)

            ' Sb (between-class scatter) not explicitly used 
            'Two different equations for Sb
            'See https://www.ics.uci.edu/~welling/teaching/KernelsICS273B/Fisher-LDA.pdf

            Console.WriteLine(VBCrLf & "End LDA binary classification demo" & VBCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function Prediction(ByVal data()() As Double, ByVal x() As Double, ByVal w()() As Double) As Integer
            ' item to predict, x, is a normal array
            ' convert array to 1-col matrix then call regular Prediction()
            Dim xm()() As Double = MatrixFromVector(x) ' x to col matrix
            Return Prediction(data, xm, w)
        End Function

        Private Shared Function Prediction(ByVal data()() As Double, ByVal x()() As Double, ByVal w()() As Double) As Integer
            ' item to predict, x, is a 1-col matrix (column vector)
            Dim [dim] As Integer = data(0).Length - 1 ' at least 2
            Dim wT()() As Double = MatrixTranspose(w) ' 1 x d
            Dim m0()() As Double = Mean(data, 0) ' d x 1
            Dim m1()() As Double = Mean(data, 1) ' d x 1
            Dim m()() As Double = MatrixAdd(m0, m1) ' d x 1
            m = MatrixProduct(m, 0.5) ' d x 1

            ' threshold constant c (distance from origin)
            Dim tc()() As Double = MatrixProduct(wT, m) ' ((1xd)(dx1) = 1 x 1
            ' distance from origin
            Dim wTx()() As Double = MatrixProduct(wT, x) ' (1xd)(dx1) = 1 x 1

            Dim adjust As Double = 0.0

            ' optional weighting by probabilities:
            'Dim n0 As Integer = 0 ' number each class
            'Dim n1 As Integer = 0
            'For i As Integer = 0 To data.Length - 1
            'Dim label As Integer = CInt(Fix(data(i)(dim)))
            '    If label = 0 Then
            '        n0 += 1
            '    Else
            '        n1 += 1
            '    End If
            'Next i
            'Dim p0 As Double = (n0 * 1.0) / data.Length
            'Dim p1 As Double = (n1 * 1.0) / data.Length
            'Dim adjust As Double = Math.Log(p0 / p1)

            If wTx(0)(0) - adjust > tc(0)(0) Then ' a bit tricky
                Return 0
            Else
                Return 1
            End If
        End Function

        Private Shared Function Mean(ByVal data()() As Double, ByVal c As Integer) As Double()()
            ' return mean of class c (0 or 1) as column vector
            Dim [dim] As Integer = data(0).Length - 1
            Dim result()() As Double = MatrixCreate([dim], 1)
            Dim ct As Integer = 0

            For i As Integer = 0 To data.Length - 1
                Dim label As Integer = CInt(Fix(data(i)([dim])))
                If label = c Then
                    For j As Integer = 0 To [dim] - 1
                        result(j)(0) += data(i)(j)
                    Next j
                    ct = ct + 1
                End If
            Next i ' i

            For i As Integer = 0 To [dim] - 1
                result(i)(0) /= ct
            Next i

            Return result
        End Function ' Mean

        Private Shared Function Discriminate(ByVal data()() As Double, ByVal unitize As Boolean) As Double()()
            ' returns the w vector as a column 
            ' calls helper ScatterWithin which calls Scatter
            Dim mean0()() As Double = Mean(data, 0)
            Dim mean1()() As Double = Mean(data, 1)
            Console.WriteLine(VBCrLf & "Class means: " & VBCrLf)
            ShowMatrix(mean0, 2, True)
            ShowMatrix(mean1, 2, True)

            Dim Sw()() As Double = ScatterWithin(data) ' sum of S0 and S1
            Console.WriteLine("The within-class combined scatter matrix Sw: " & VBCrLf)
            ShowMatrix(Sw, 4, True)

            Dim SwInv()() As Double = MatrixInverse(Sw)

            Dim diff()() As Double = MatrixSubtract(mean0, mean1)
            Dim w()() As Double = MatrixProduct(SwInv, diff)

            If unitize = True Then
                Return LinearDiscriminateProgram.Unitize(w)
            Else
                Return w
            End If
        End Function

        Private Shared Function ScatterWithin(ByVal data()() As Double) As Double()()
            ' Sw = within class scatter = S0 + S1
            Dim S0()() As Double = Scatter(data, 0)
            Dim S1()() As Double = Scatter(data, 1)
            Console.WriteLine("Scatter matrices S0 S1: " & VBCrLf)
            ShowMatrix(S0, 4, True)
            ShowMatrix(S1, 4, True)
            Dim Sw()() As Double = MatrixAdd(S0, S1)
            Return Sw
        End Function

        Private Shared Function Scatter(ByVal data()() As Double, ByVal c As Integer) As Double()()
            ' scatter matrix of class c
            ' [Sw (within class) is sum of class scatters]
            Dim [dim] As Integer = data(0).Length - 1
            Dim mean()() As Double = LinearDiscriminateProgram.Mean(data, c) ' mean as col vector
            Dim result()() As Double = MatrixCreate([dim], [dim]) ' d x d
            For i As Integer = 0 To data.Length - 1
                Dim label As Integer = CInt(Fix(data(i)([dim]))) ' 0 or 1
                If label = c Then
                    Dim x()() As Double = MatrixCreate([dim], 1) ' d x 1
                    For j As Integer = 0 To [dim] - 1
                        x(j)(0) = data(i)(j)
                    Next j
                    Dim diff()() As Double = MatrixSubtract(x, mean) ' d x 1
                    Dim diffT()() As Double = MatrixTranspose(diff) ' 1 x d
                    Dim prod()() As Double = MatrixProduct(diff, diffT) ' d x d

                    result = MatrixAdd(result, prod) ' accumulate
                End If
            Next i ' i
            Return result
        End Function ' Scatter


        Private Shared Function Unitize(ByVal vector()() As Double) As Double()()
            ' return column vector scaled to unit length
            If vector(0).Length <> 1 Then
                Throw New Exception("Not column vector")
            End If

            Dim len As Integer = vector.Length
            Dim result()() As Double = MatrixCreate(len, 1)
            Dim sum As Double = 0.0
            For i As Integer = 0 To len - 1
                sum += vector(i)(0) * vector(i)(0)
            Next i
            sum = Math.Sqrt(sum) ' check if 0
            For i As Integer = 0 To len - 1
                result(i)(0) = vector(i)(0) / sum
            Next i
            Return result
        End Function

        ' ============================================

        'Shared Function Covariance(ByVal data()() As Double, ByVal c As Integer, ByVal sample As Boolean) As Double()()
        '    ' covariance is essentially Scatter divided by number items
        '    Dim n As Integer = 0
        '    Dim [dim] As Integer = data(0).Length - 1
        '    For i As Integer = 0 To data.Length - 1
        '        Dim label As Integer = CInt(Fix(data(i)([dim])))
        '        If label = c Then
        '            n += 1
        '        End If
        '    Next i
        '    Dim s()() As Double = Scatter(data, c)
        '    Dim result()() As Double = MatrixDuplicate(s)
        '    For i As Integer = 0 To [dim] - 1
        '        For j As Integer = 0 To [dim] - 1
        '            If sample = True Then
        '                result(i)(j) /= n - 1
        '            Else
        '                result(i)(j) /= n
        '            End If
        '        Next j
        '    Next i
        '    Return result
        'End Function ' Covariance

        'Shared Function Pooled(ByVal data()() As Double, ByVal sample As Boolean) As Double()()
        '    ' pooled covariance: (n1*C1 + n2*C2) / (n1 + n2)
        '    Dim n0 As Integer = 0
        '    Dim n1 As Integer = 0
        '    Dim [dim] As Integer = data(0).Length - 1
        '    For i As Integer = 0 To data.Length - 1
        '        Dim label As Integer = CInt(Fix(data(i)([dim]))) ' 0 or 1
        '        If label = 0 Then
        '            n0 += 1
        '        Else
        '            n1 += 1
        '        End If
        '    Next i

        '    Dim C0()() As Double = Covariance(data, 0, sample)
        '    Dim C1()() As Double = Covariance(data, 1, sample)

        '    Dim result()() As Double = MatrixCreate([dim], [dim])
        '    For i As Integer = 0 To [dim] - 1
        '        For j As Integer = 0 To [dim] - 1
        '            result(i)(j) = (n0 * C0(i)(j) + n1 * C1(i)(j)) / (n0 + n1)
        '        Next j
        '    Next i
        '    Return result
        'End Function ' Pooled

        ' -------------------------------------------------------------
        ' Matrix methods below
        ' -------------------------------------------------------------

        Private Shared Function MatrixInverse(ByVal matrix()() As Double) As Double()()
            Dim n As Integer = matrix.Length
            Dim result()() As Double = MatrixDuplicate(matrix)

            Dim perm() As Integer = Nothing
            Dim toggle As Integer
            Dim lum()() As Double = MatrixDecompose(matrix, perm, toggle)
            If lum Is Nothing Then
                Throw New Exception("Unable to compute inverse")
            End If

            Dim b(n - 1) As Double
            For i As Integer = 0 To n - 1
                For j As Integer = 0 To n - 1
                    If i = perm(j) Then
                        b(j) = 1.0
                    Else
                        b(j) = 0.0
                    End If
                Next j

                Dim x() As Double = HelperSolve(lum, b) ' use decomposition

                For j As Integer = 0 To n - 1
                    result(j)(i) = x(j)
                Next j
            Next i
            Return result
        End Function

        ' -------------------------------------------------------------

        Private Shared Function HelperSolve(ByVal luMatrix()() As Double, ByVal b() As Double) As Double()
            ' before calling this helper, permute b using the perm array
            ' from MatrixDecompose that generated luMatrix
            Dim n As Integer = luMatrix.Length
            Dim x(n - 1) As Double
            b.CopyTo(x, 0)

            For i As Integer = 1 To n - 1
                Dim sum As Double = x(i)
                For j As Integer = 0 To i - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum
            Next i

            x(n - 1) /= luMatrix(n - 1)(n - 1)
            For i As Integer = n - 2 To 0 Step -1
                Dim sum As Double = x(i)
                For j As Integer = i + 1 To n - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum / luMatrix(i)(i)
            Next i

            Return x
        End Function

        ' -------------------------------------------------------------

        Private Shared Function MatrixDuplicate(ByVal matrix()() As Double) As Double()()
            ' allocates/creates a duplicate of a matrix
            Dim result()() As Double = MatrixCreate(matrix.Length, matrix(0).Length)
            For i As Integer = 0 To matrix.Length - 1 ' copy the values
                For j As Integer = 0 To matrix(i).Length - 1
                    result(i)(j) = matrix(i)(j)
                Next j
            Next i
            Return result
        End Function

        ' -------------------------------------------------------------

        Private Shared Function MatrixDecompose(ByVal matrix()() As Double, <System.Runtime.InteropServices.Out()> ByRef perm() As Integer, <System.Runtime.InteropServices.Out()> ByRef toggle As Integer) As Double()()
            ' Doolittle LUP decomposition with partial pivoting.
            ' returns: result is L (with 1s on diagonal) and U;
            ' perm holds row permutations; toggle is +1 or -1 (even or odd)
            Dim rows As Integer = matrix.Length
            Dim cols As Integer = matrix(0).Length
            If rows <> cols Then
                Throw New Exception("Non-square mattrix")
            End If

            Dim n As Integer = rows ' convenience

            Dim result()() As Double = MatrixDuplicate(matrix)

            perm = New Integer(n - 1) {} ' set up row permutation result
            For i As Integer = 0 To n - 1
                perm(i) = i
            Next i

            toggle = 1 ' toggle tracks row swaps

            For j As Integer = 0 To n - 2 ' each column
                Dim colMax As Double = Math.Abs(result(j)(j))
                Dim pRow As Integer = j
                'For i As Integer = j + 1 To n - 1 ' deprecated
                '    If result(i)(j) > colMax Then
                '        colMax = result(i)(j)
                '        pRow = i
                '    End If
                'Next i

                For i As Integer = j + 1 To n - 1 ' reader Matt V needed this:
                    If Math.Abs(result(i)(j)) > colMax Then
                        colMax = Math.Abs(result(i)(j))
                        pRow = i
                    End If
                Next i
                ' Not sure if this approach is needed always, or not.

                If pRow <> j Then ' if largest value not on pivot, swap rows
                    Dim rowPtr() As Double = result(pRow)
                    result(pRow) = result(j)
                    result(j) = rowPtr

                    Dim tmp As Integer = perm(pRow) ' and swap perm info
                    perm(pRow) = perm(j)
                    perm(j) = tmp

                    toggle = -toggle ' adjust the row-swap toggle
                End If

                ' -------------------------------------------------------------
                ' This part added later (not in original code) 
                ' and replaces the 'return null' below.
                ' if there is a 0 on the diagonal, find a good row 
                ' from i = j+1 down that doesn't have
                ' a 0 in column j, and swap that good row with row j

                If result(j)(j) = 0.0 Then
                    ' find a good row to swap
                    Dim goodRow As Integer = -1
                    For row As Integer = j + 1 To n - 1
                        If result(row)(j) <> 0.0 Then
                            goodRow = row
                        End If
                    Next row

                    If goodRow = -1 Then
                        Throw New Exception("Cannot use Doolittle's method")
                    End If

                    ' swap rows so 0.0 no longer on diagonal
                    Dim rowPtr() As Double = result(goodRow)
                    result(goodRow) = result(j)
                    result(j) = rowPtr

                    Dim tmp As Integer = perm(goodRow) ' and swap perm info
                    perm(goodRow) = perm(j)
                    perm(j) = tmp

                    toggle = -toggle ' adjust the row-swap toggle
                End If
                ' -------------------------------------------------------------

                'If Math.Abs(result(j)(j)) < 1.0E-20 Then ' deprecated
                '    Return Nothing ' consider a throw
                'End If

                For i As Integer = j + 1 To n - 1
                    result(i)(j) /= result(j)(j)
                    For k As Integer = j + 1 To n - 1
                        result(i)(k) -= result(i)(j) * result(j)(k)
                    Next k
                Next i

            Next j ' main j column loop

            Return result
        End Function ' MatrixDecompose

        ' -------------------------------------------------------------

        Private Shared Function MatrixProduct(ByVal matrixA()() As Double, ByVal matrixB()() As Double) As Double()()
            Dim aRows As Integer = matrixA.Length
            Dim aCols As Integer = matrixA(0).Length
            Dim bRows As Integer = matrixB.Length
            Dim bCols As Integer = matrixB(0).Length
            If aCols <> bRows Then
                Throw New Exception("Non-conformable matrices in MatrixProduct")
            End If

            Dim result()() As Double = MatrixCreate(aRows, bCols)

            For i As Integer = 0 To aRows - 1 ' each row of A
                For j As Integer = 0 To bCols - 1 ' each col of B
                    For k As Integer = 0 To aCols - 1 ' could use k < bRows
                        result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
                    Next k
                Next j
            Next i

            'Parallel.For(0, aRows, Sub(i)
            '                           For j As Integer = 0 To bCols - 1
            '                               For k As Integer = 0 To aCols - 1
            '                                   result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
            '                               Next
            '                           Next
            '                       End Sub)

            Return result
        End Function

        Private Shared Function MatrixProduct(ByVal m()() As Double, ByVal x As Double) As Double()()
            ' multiple all cells in m by scalar x
            Dim result()() As Double = MatrixDuplicate(m) ' copy
            For i As Integer = 0 To result.Length - 1
                For j As Integer = 0 To result(i).Length - 1
                    result(i)(j) *= x
                Next j
            Next i
            Return result
        End Function

        Private Shared Function MatrixAdd(ByVal a()() As Double, ByVal b()() As Double) As Double()()
            ' return a-b
            Dim rows As Integer = a.Length
            Dim cols As Integer = a(0).Length
            Dim result()() As Double = MatrixCreate(rows, cols)
            For i As Integer = 0 To rows - 1
                For j As Integer = 0 To cols - 1
                    result(i)(j) = a(i)(j) + b(i)(j)
                Next j
            Next i
            Return result
        End Function

        Private Shared Function MatrixSubtract(ByVal a()() As Double, ByVal b()() As Double) As Double()()
            ' return a-b
            Dim rows As Integer = a.Length
            Dim cols As Integer = a(0).Length
            Dim result()() As Double = MatrixCreate(rows, cols)
            For i As Integer = 0 To rows - 1
                For j As Integer = 0 To cols - 1
                    result(i)(j) = a(i)(j) - b(i)(j)
                Next j
            Next i
            Return result
        End Function

        Private Shared Function MatrixTranspose(ByVal matrix()() As Double) As Double()()
            Dim rows As Integer = matrix.Length
            Dim cols As Integer = matrix(0).Length
            Dim result()() As Double = MatrixCreate(cols, rows) ' note indexing
            For i As Integer = 0 To rows - 1
                For j As Integer = 0 To cols - 1
                    result(j)(i) = matrix(i)(j)
                Next j
            Next i
            Return result
        End Function ' TransposeMatrix

        Private Shared Function MatrixFromVector(ByVal vector() As Double) As Double()()
            ' return a column vector-matrix
            Dim len As Integer = vector.Length
            Dim result()() As Double = MatrixCreate(len, 1) ' 1 colum
            For i As Integer = 0 To len - 1
                result(i)(0) = vector(i)
            Next i
            Return result
        End Function

        Private Shared Function MatrixCreate(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            ' allocates/creates a matrix initialized to all 0.0
            ' do error checking here
            Dim result(rows - 1)() As Double
            For i As Integer = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Private Shared Sub ShowMatrix(ByVal m()() As Double, ByVal dec As Integer, ByVal newl As Boolean)
            For i As Integer = 0 To m.Length - 1
                For j As Integer = 0 To m(i).Length - 1
                    If m(i)(j) >= 0.0 Then ' '+'
                        Console.Write(" ")
                    End If
                    Console.Write(m(i)(j).ToString("F" & dec) & "  ")
                Next j
                Console.WriteLine("")
            Next i
            If newl Then
                Console.WriteLine("")
            End If
        End Sub

        Private Shared Sub ShowVector(ByVal v() As Double, ByVal dec As Integer)
            For i As Integer = 0 To v.Length - 1
                Console.Write(v(i).ToString("F" & dec) & "  ")
            Next i
            Console.WriteLine("")
        End Sub

    End Class ' Program clss

End Namespace ' ns


' R language analysis for validation of VB program
'> library(MASS)
'>
'> data1 = read.table(file.choose(), header=T, sep=',')
'>
'> data1
'  Age Income Politic
'1   1      4       A
'2   2      2       A
'3   3      3       A
'4   6      6       B
'5   7      7       B
'6   8      9       B
'7   8      7       B
'8   9      8       B
'>
'> data1.lda <- lda(formula = Politic ~ Age + Income, data = data1)
'>
'> data1.lda

'Prior probabilities of groups:
'    A     B
'0.375 0.625

'Group means:
'  Age Income
'A 2.0    3.0
'B 7.6    7.4

'Coefficients of linear discriminants:
'             LD1
'Age    0.6859669
'Income 0.3919811
'>

' Show the VB w vector is equivalent to the R result
'> v = c(0.6859669, 0.3919811)
'> len = sqrt(v[1]^2 + v[2]^2)
'> w = v / len
'> w
'[1] 0.8682431 0.4961390
'>

' Make a prediction using R
'> unknown <- data.frame(Age=4,Income=5,Politic='?')
'> predict(data1.lda, unknown)
'          A          B
'1 0.9516662 0.04833379

' a few LDA resources
'See http://www.cbcb.umd.edu/~hcorrada/PracticalML/pdf/lectures/classification.pdf
'See http://sites.stat.psu.edu/~jiali/course/stat597e/notes2/lda.pdf
'See http://www.saedsayad.com/lda.htm